import zmq, time
import multiprocessing as mp
import pystreaming.video.interface as intf
from pystreaming.listlib.circularlist import CircularList, Empty


# Note. It might be better to implement this as a asyncio system?


def dist_ps(shutdown, infd, endpt, rcvhwm, tracks):
    print(f"Start Distributing to {endpt}")
    context = zmq.Context()
    collector = context.socket(zmq.PULL)
    collector.setsockopt(zmq.RCVHWM, rcvhwm)
    collector.bind(infd)
    distributor = context.socket(zmq.REP)
    distributor.bind(endpt)
    if tracks is None:  # default track is "none"
        tracks = ["none"]
    queues = {track: CircularList() for track in tracks}
    while not shutdown.is_set():
        time.sleep(0.001)  # 1000 cycles/sec ~> 4-6 streams at once.
        if collector.poll(0):  # returns 0 if no event, something else if there is
            buf, idx = intf.recv_buf_idx(collector, flags=zmq.NOBLOCK)
            for fqueue in queues.values():
                fqueue.push((buf, idx))  # add to buf queue
        if distributor.poll(0):  # got frame req
            track = distributor.recv().decode()
            try:
                buf, idx = queues[track].pop()
                intf.send_buf_idx(distributor, buf, idx, flags=zmq.NOBLOCK)
            except (KeyError, Empty):  # no frames available or wrong track selected
                intf.send_buf_idx(distributor, b"nil", -1, flags=zmq.NOBLOCK)

    print("Distributor stopped")


class Distributor:
    infd = "ipc:///tmp/encout"
    rcvhwm = 30

    def __init__(self, endpoint, tracks=None):
        self.shutdown = mp.Event()
        self.ps = None
        self.psargs = (self.shutdown, self.infd, endpoint, self.rcvhwm, tracks)

    def start(self):
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning Distributor")

        self.ps = mp.Process(target=dist_ps, args=self.psargs)
        self.ps.daemon = True
        self.ps.start()

    def stop(self):
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped Distributor")

        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()



from turbojpeg import TurboJPEG, TJSAMP_420, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import numpy as np
import zmq, time
import multiprocessing as mp
from functools import partial
import pystreaming.video.interface as intf


def enc_ps(shutdown, infd, outfd, rcvhwm, outhwm):
    print("Starting Encoder")
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)
    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, outhwm)
    out.connect(outfd)
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    encoder = partial(
        TurboJPEG().encode, quality=70, jpeg_subsample=TJSAMP_420, flags=TJFLAG_FASTDCT
    )
    while not shutdown.is_set():
        time.sleep(0.01)  # 100 cycles/sec
        if poller.poll(0):
            try:
                arr, idx = intf.recv_ndarray_idx(socket, flags=zmq.NOBLOCK)
                intf.send_buf_idx(out, encoder(arr), idx, flags=zmq.NOBLOCK)
            except zmq.error.Again:
                # ignore send misses to distributor.
                # Should not happen if there is a distributor.
                # Could implement a signal that there is a distributor. Not strictly necessary tho.
                pass
    print("Stopping Encoder")


class Encoder:
    infd = "ipc:///tmp/encin"
    outfd = "ipc:///tmp/encout"
    tellhwm = 30
    rcvhwm = outhwm = 10

    def __init__(self, context, procs=2):
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.psargs = (self.shutdown, self.infd, self.outfd, self.rcvhwm, self.outhwm)
        self.workers = []
        self.idx = 0

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.setsockopt(zmq.SNDHWM, self.tellhwm)
        self.sender.bind(self.infd)

    def send(self, frame):
        if self.workers == []:
            raise RuntimeError("Tried to send frame to stopped Encoder")
        try:
            intf.send_ndarray_idx(self.sender, frame, self.idx, flags=zmq.NOBLOCK)
            self.idx += 1
        except zmq.error.Again:
            raise RuntimeError("Worker processes are not processing frames fast enough")
        # change behavior to:
        # IF working, but slow, print a warning
        # If not responding, only then raise runtime error

    def start(self):
        if self.workers != []:
            raise RuntimeError("Tried to start running Encoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=enc_ps, args=self.psargs))

        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(1)  # Allow workers to spin up

    def stop(self):
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Encoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()


from turbojpeg import TurboJPEG, TJSAMP_420, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import cv2
import numpy as np
import zmq, time, asyncio, zmq.asyncio
import multiprocessing as mp
from functools import partial
from ..listlib.circularlist import CircularList


def recv_ndarray_idx(socket):  # returns frame, idx
    md = socket.recv_json()
    msg = socket.recv(copy=False)
    buf = memoryview(msg)
    A = np.frombuffer(buf, dtype=md["dtype"])
    return A.reshape(md["shape"]), md["idx"]


def send_ndarray_idx(arr, socket, idx):
    md = dict(
        dtype=str(arr.dtype),
        shape=arr.shape,
        idx=idx,
    )
    socket.send_json(md, zmq.SNDMORE)
    socket.send(arr, copy=False)


def enc_ps(shutdown, infd, outfd):
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect(infd)
    out = context.socket(zmq.PUSH)
    out.connect(outfd)
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    encoder = partial(
        TurboJPEG().encode, quality=70, jpeg_subsample=TJSAMP_420, flags=TJFLAG_FASTDCT
    )
    while not shutdown.is_set():
        if poller.poll(0):
            arr, idx = recv_ndarray_idx(socket)
            buf = encoder(arr)
            out.send(buf, copy=False, flags=zmq.SNDMORE)
            out.send_pyobj(idx)
            print(f"ENC: idx={idx}")


def dec_ps(shutdown, infd, outfd):
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.connect(infd)
    out = context.socket(zmq.PUSH)
    out.connect(outfd)
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    decoder = partial(TurboJPEG().decode, flags=(TJFLAG_FASTDCT + TJFLAG_FASTUPSAMPLE))

    while not shutdown.is_set():
        if poller.poll(0):
            buf = socket.recv()
            idx = socket.recv_pyobj()
            frame = decoder(buf)
            send_ndarray_idx(frame, out, idx)


def distributor(shutdown, infd, endpoint):
    context = zmq.Context()
    collector = context.socket(zmq.PULL)
    collector.bind(infd)
    distributor = context.socket(zmq.REP)
    distributor.bind(endpoint)
    queue = CircularList()
    queues = {}
    while not shutdown.is_set():
        if collector.poll(0):  # returns 0 if no event, something else if there is
            buf = collector.recv()  # get buf
            idx = collector.recv_pyobj()  # then get idx
            queue.push((buf, idx))  # add to queue
        if distributor.poll(0):  # got frame req
            _ = distributor.recv()
            package = queue.pop()
            buf, idx = (b"nil", -1) if package is None else package
            distributor.send(
                buf, copy=False, flags=zmq.SNDMORE
            )  # send out buf, the idx
            distributor.send_pyobj(idx)


async def aioreq(context, source, shutdown, drain):
    socket = context.socket(zmq.REQ)
    socket.connect(f"{source}")
    print("Started a frame request async thread")
    while not shutdown.is_set():
        await socket.send(b"TRACK1")
        buf = await socket.recv()
        idx = await socket.recv_pyobj()
        if idx == -1:
            continue
        print(f"AIO: Dropped in drain: {idx}")
        await drain.send(buf, copy=False, flags=zmq.SNDMORE)
        await drain.send_pyobj(idx)


def aiomain(source, outfd, procs, shutdown):
    context = zmq.asyncio.Context()
    drain = context.socket(zmq.PUSH)
    drain.bind(outfd)
    args = [aioreq(context, source, shutdown, drain) for _ in range(procs)]
    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(*args))


class UltraJPGEnc:
    infd = "ipc:///tmp/encin"
    outfd = "ipc:///tmp/encout"

    def __init__(self, context, procs=2):
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.workers = []
        self.idx = 0

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.bind(self.infd)

    def tell(self, frame):
        send_ndarray_idx(frame, self.sender, self.idx)
        self.idx += 1

    def start_workers(self):
        if self.workers != []:
            raise RuntimeError(
                "Warning: tried to start running jpg encoder bank... Ignoring"
            )
        self.workers = [
            mp.Process(target=enc_ps, args=(self.shutdown, self.infd, self.outfd))
            for _ in range(self.procs)
        ]
        for ps in self.workers:
            ps.daemon = True
            ps.start()

    def stop_workers(self):
        if self.workers == []:
            print("Warning: tried to stop stopped jpg encoder bank... Ignoring")
            return

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()


class AIOREQ:
    outfd = "ipc:///tmp/decin"

    def __init__(self, source, procs=3):
        self.source, self.procs = source, procs
        self.shutdown = mp.Event()
        self.ps = None

    def start(self):
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning AIOREQ obj")

        self.ps = mp.Process(
            target=aiomain,
            args=(self.source, self.outfd, self.procs, self.shutdown),
        )
        self.ps.daemon = True
        self.ps.start()

    def stop(self):
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped AIOREQ obj")
        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()


class UltraJPGDec:
    infd = "ipc:///tmp/decin"
    outfd = "ipc:///tmp/decout"

    def __init__(self, context, procs=2):
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.workers = []

        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.bind(self.outfd)

    def hear(self):  # returns frame, idx
        return recv_ndarray_idx(self.receiver)

    def start_workers(self):
        if self.workers != []:
            print("Warning: tried to start a running jpg decoder bank... Ignoring")
            return

        for _ in range(self.procs):
            self.workers.append(
                mp.Process(target=dec_ps, args=(self.shutdown, self.infd, self.outfd))
            )
        for ps in self.workers:
            ps.daemon = True
            ps.start()

    def stop_workers(self):
        if self.workers == []:
            print("Warning: tried to stop a stopped jpg decoder bank... Ignoring")
            return

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()
        self.running = False


class Sequencer:
    infd1


class VideoPlayer:
    infd = "ipc:///tmp/sequenced"

    def play(self, context):
        socket = context.socket(zmq.SUB)
        socket.connect(self.infd)
        while True:
            id, frame, _ = socket.recv
            print(f"\r frame: {idx}", end="")
            cv2.imshow("mpjpeg.VideoPlayer", frame)
            if cv2.waitKey(1):
                break
        cv2.destroyAllWindows()


class VideoWriter:
    ...


# STREAM # cam =frame> ENC =buf,idx> DIS =buf,idx>

# WORKER # =buf,idx> AIOREQ =buf,idx> DEC =frame,idx> func =res,idx> PUSH =res,idx>

# COLLEX # =res,idx> PULL =(resout)res,idx>
# COLLEX # =buf,idx> AIOREQ =buf,idx> DEC =(decout)frame,idx>
# COLLEX # =res,frame,idx> SEQ =(sequenced)res,frame,idx>
# COLLEX # =res,frame,idx> Filewriter
# COLLEX # =res,frame,idx> VideoPlayer


def encdistmain():
    frame = cv2.imread("pystreaming/pycodec/test_imgs/1280x720.jpg")
    context = zmq.Context()
    bank = UltraJPGEnc(context)
    dist = Distributor("tcp://127.0.0.1:5553")
    bank.start_workers()
    dist.start()
    for i in range(100):
        print(i)
        time.sleep(0.5)
        bank.tell(frame)
    print("call1")
    bank.stop_workers()
    print("call2")
    dist.stop()


if __name__ == "__main__":
    encdistmain()
    # context = zmq.Context()
    # bank = UltraJPGEnc(context, procs=2)
    # bank.start_workers()
    # cap = cv2.VideoCapture(0)
    # cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"MJPG"))
    # cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 1080)
    # cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)
    # start = time.time()
    # for i in range(1000):
    #     print(i)
    #     ret, frame = cap.read()
    #     bank.tell(frame)
    #     if i > 4:
    #         buf = bank.hear()
    # print(f"Done in {1000/(time.time() - start)} FPS")

    # bank.stop_workers()



import numpy as np
import zmq

def send_ndarray_idx(socket, arr, idx, flags=0):
    """Send a numpy array and an index.

    Args:
        socket (zmq.Context.socket): Socket through which to send data.
        arr (numpy.ndarray): Numpy array to send.
        idx (int): Index of the sent object.
        flags (int, optional): Zmq flags to execute with 
            (only zmq.NOBLOCK or zmq.SNDMORE). Defaults to 0.
    """
    md = dict(
        dtype=str(arr.dtype),
        shape=arr.shape,
        idx=idx,
    )
    socket.send_json(md, flags=zmq.SNDMORE|flags)
    socket.send(arr, copy=False, flags=flags)

def recv_ndarray_idx(socket, flags=0):
    """Receive a numpy array and an index.

    Args:
        socket (zmq.Context.socket): Socket through which to receive data.
        flags (int, optional): Zmq flags to execute with 
            (only zmq.NOBLOCK). Defaults to 0.

    Returns:
        [tuple(numpy.ndarray, int)]: Numpy array and index received.
    """
    md = socket.recv_json(flags=flags)
    msg = socket.recv(copy=False, flags=flags)
    buf = memoryview(msg)
    A = np.frombuffer(buf, dtype=md["dtype"])
    return A.reshape(md["shape"]), md["idx"]


def send_buf_idx(socket, buf, idx, flags=0):
    """Send a buffer and an index.

    Args:
        socket (zmq.Context.socket): Socket through which to send data.
        buf (bytes): Buffer to send.
        idx (int): Index of the sent object.
        flags (int, optional): Zmq flags to execute with 
            (only zmq.NOBLOCK or zmq.SNDMORE). Defaults to 0.
    """
    socket.send(buf, copy=False, flags=zmq.SNDMORE|flags)
    socket.send_pyobj(idx, flags=flags)

def recv_buf_idx(socket, flags=0):
    """Receive a buffer and an index.

    Args:
        socket (zmq.Context.socket): Socket through which to receive data.
        flags (int, optional): Zmq flags to execute with 
            (only zmq.NOBLOCK). Defaults to 0.

    Returns:
        [tuple(bytes, int)]: Buffer and index received.
    """
    buf = socket.recv(copy=False, flags=flags)
    idx = socket.recv_pyobj(flags=flags)
    return buf, idx


import zmq, zmq.asyncio
import asyncio, time
import multiprocessing as mp

async def aioreq(context, source, shutdown, drain):
    socket = context.socket(zmq.REQ)
    socket.connect(source)
    print("Started a frame request async thread")
    while not shutdown.is_set():
        await socket.send(b"TRACK1")
        buf = await socket.recv()
        idx = await socket.recv_pyobj()
        if idx == -1:
            continue
        print(f"AIO: Dropped in drain: {idx}")
        await drain.send(buf, copy=False, flags=zmq.SNDMORE)
        await drain.send_pyobj(idx)

async def workz(shutdown):
    while True:
        print("Work...")
        await asyncio.sleep(1)
async def stopcheck(shutdown):
    while True:
        print("SLEEP")
        await asyncio.sleep(5)
        print("hehexd RAISE")
        raise StopAsyncIteration

def aiomain(shutdown):
    args = [stopcheck(1), workz(1), workz(2)]
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(asyncio.gather(*args))
    except StopAsyncIteration as e:
        loop.stop()
        print("STOPPED DUE TO RAISED ERROR")


def main():
    time.sleep(1)
    ps = mp.Process(
            target=aiomain,
            args=(1,)
    )
    ps.daemon = True
    ps.start()
    ps.join()
    
if __name__ == "__main__":
    main()


class AIOREQ:
    outfd = "ipc:///tmp/decin"

    def __init__(self, source, procs=3):
        self.source, self.procs = source, procs
        self.shutdown = mp.Event()
        self.ps = None

    def start(self):
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning AIOREQ obj")

        self.ps = mp.Process(
            target=aiomain,
            args=(self.source, self.outfd, self.procs, self.shutdown),
        )
        self.ps.daemon = True
        self.ps.start()

    def stop(self):
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped AIOREQ obj")
        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()


import time, zmq
from pystreaming import Requester, Encoder, Distributor, Decoder

def main():
    context = zmq.Context()
    x = Requester("tcp://127.0.0.1:5555")
    x.start()
    time.sleep(1)
    x.stop()

    time.sleep(1)
    
    x = Encoder(context)
    x.start()
    time.sleep(1)
    x.stop()
    
    x = Distributor("tcp://*:5555")
    x.start()
    time.sleep(1)
    x.stop()
    
    x = Decoder(context)
    x.start()
    time.sleep(1)
    x.stop()
    
if __name__ == "__main__":
    main()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pystreaming",
    version=open("VERSION.txt").read().strip(),
    author="Joseph Li",
    author_email="jxli@andrew.cmu.edu",
    description="Video + Audio streaming done with ZMQ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/joseph-x-li/pystreaming",
    packages=setuptools.find_packages(),
    install_requires=["numpy"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)